; 
;
;               HASH LIBRARY                               by Nmeth Gbor
;
;
; 

    .486                       ; create 32 bit code
    .model flat, stdcall       ; 32 bit memory model
    option casemap :none       ; case sensitive

    include     \masm32\include\windows.inc
    include     \masm32\include\kernel32.inc
    include     \masm32\macros\macros.asm

; needed for debugging
    include     \masm32\include\masm32.inc
    include     \masm32\include\debug.inc

    includelib  \masm32\lib\kernel32.lib

; needed for debugging
    includelib  \masm32\lib\debug.lib
    includelib  \masm32\lib\masm32.lib

; 
;
; FUNCTION PROTOTYPES
;
; HASH1_init	PROTO :DWORD,:DWORD,:DWORD,:DWORD,:DWORD
; HASH1_destroy	PROTO :DWORD
; HASH1_store	PROTO :DWORD
; HASH1_search	PROTO :DWORD
;
; 
;
; HASH1 STRUCTURE for constant size records
;

HASH1_STRUCT    STRUCT

                START       dd ?        ; start of hash array
                rSIZE       dd ?        ; size of one record
                COUNT       dd ?        ; max entry number
                FREE        dd ?        ; number of free slots  
                GETKEY      dd ?        ; pointer of hashing function
                PROBE       dd ?        ; pointer of probe function
                MAXPROBE    dd ?        ; maximum number of probes
                
HASH1_STRUCT    ENDS

; 

.data
; needed for debugging

        PUBLIC collision
            collision   dd 0            ; number of collisions during storing

; 

.code
; 
;
; DEFAULT HASH FUNCTION
;
; When the user does not assign a hash function in HASH1_init this is used
;
; IN:   eax - input data
;       edx - address of hash structure
;
; OUT:  ebx - key/index
;
; -------------------------------------------------------------------------
HASH1_DEF_FUNC1 PROC USES eax ecx edx
align 4
                ASSUME edx:PTR HASH1_STRUCT
                mov     ebx,[edx].COUNT                 ; get hash modulus (size)
                ASSUME edx:NOTHING

; this is a slightly modified version of rand proc in nrand.asm
    xor edx, edx
    mov ecx, 127773
    div ecx
    mov ecx, eax
    mov eax, 16807
    mul edx
    mov edx, ecx
    mov ecx, eax
    mov eax, 2836
    mul edx
    sub ecx, eax
    xor edx, edx
    mov eax, ecx

    div ebx
    mov ebx,edx                                         ; output is a random number

;PrintHex ebx,'key1'
                ret

HASH1_DEF_FUNC1 ENDP
; 
;
; DEFAULT PROBE FUNCTION (linear probe)
;
; When the user does not assign a probe function in HASH1_init this is used
;
; IN:   ebx - input key/index
;       edx - address of hash structure
;
; OUT:  ebx - key/index
;
; -------------------------------------------------------------------------
HASH1_DEF_FUNC2 PROC USES eax
align 4
                or      ebx,ebx
                jnz  @_1
                ASSUME edx:PTR HASH1_STRUCT
                mov     ebx,[edx].COUNT
                ASSUME edx:NOTHING
@_1:
                sub     ebx,1
;PrintHex ebx,'key2'

; needed for debugging
                add dword ptr collision,1               ; increase the number of collisions
                ret

HASH1_DEF_FUNC2 ENDP

; 
;
; Initialize HASH utility
;
; IN:   hd  - address of structure for hash data
;       hn  - max. number of entries (modulus/size)
;       rs  - record size in DWORDs
;       hf1 - hash function address
;       hf2 - probe function address
;
; OUT:  fail    eax - NULL
;       succes  eax - address of hash store area
;
; -------------------------------------------------------------------------
HASH1_init      PROC USES ebx edx esi hd:DWORD,hn:DWORD,rs:DWORD,hf1:DWORD,hf2:DWORD

                mov     esi,hd
                ASSUME esi:PTR HASH1_STRUCT
                mov     eax,hf1
                mov     ebx,hf2
                cmp     eax,NULL
                jnz  @_1
                lea     eax,HASH1_DEF_FUNC1
                lea     ebx,HASH1_DEF_FUNC2
@_1:
                mov     [esi].GETKEY,eax
                mov     [esi].PROBE,ebx
                mov     DWORD PTR [esi].MAXPROBE,0
                push    edx
                xor     edx,edx
                mov     ebx,rs                          ; size of record
                mov     [esi].rSIZE,ebx
                add     ebx,1                           ; size of key
                shl     ebx,2                           ; *4, size is in DWORDs
                mov     eax,hn
                mov     [esi].COUNT,eax
                mov     [esi].FREE,eax
                mul     ebx
;PrintHex eax,'hash size'
                invoke GlobalAlloc,GPTR,eax
                mov     [esi].START,eax
                ASSUME esi:NOTHING
                ret

HASH1_init      ENDP

; 
;
; Destroys HASH utility
;
; IN: hd - pointer to a HASH structure
;
; -------------------------------------------------------------------------
HASH1_destroy   PROC USES esi hd:DWORD

                mov     esi,hd
                ASSUME esi:PTR HASH1_STRUCT
                xor     eax,eax
                mov     [esi].COUNT,eax
                mov     [esi].GETKEY,eax
                invoke HeapFree,FUNC(GetProcessHeap),0,[esi].START
                mov     [esi].START,eax
                ASSUME esi:NOTHING
                ret

HASH1_destroy   ENDP

; 

; 
;
; Stores a data in HASH
;
; IN:     hd - pointer to a HASH structure
;        eax - key of record
;     lpData - PTR to record to store
;
; OUT: ebx - key/index
;
; -------------------------------------------------------------------------
HASH1_store     PROC USES eax ecx edx esi edi hd:DWORD,lpData:DWORD
align 4
                mov     edx,hd
                ASSUME edx:PTR HASH1_STRUCT
                cmp     [edx].FREE,0
                jz  @_4
                xor     ecx,ecx                         ; init number of probes with 0
                mov     esi,[edx].START

                call    DWORD PTR [edx].GETKEY          ; hash function: ebx - key/index
@_1:
;PrintHex ebx,'hashkey'
                push    eax
                push    edx
                mov     eax,[edx].rSIZE                 ; size of record
                add     eax,1                           ; size of key
                shl     eax,2                           ; size in DWORDs
                mul     ebx
                pop     edx
                mov     edi,eax
                pop     eax

                cmp     DWORD PTR [esi+edi],0           ; is the slot empty?
                jz @_2

                add     ecx,1                           ; increase number of probes
                call    DWORD PTR [edx].PROBE           ; hash function2 >> ebx: key/index
                jmp @_1
@_2:
                add     esi,edi
                mov     [esi],eax                       ; store the key
;PrintHex eax,'key'
                mov     ebx,lpData
                push    ecx
                mov     ecx,[edx].rSIZE
                mov     eax,lpData
align 4
@_21:
                mov     eax,[ebx+ecx*4-4]               ; store the...
                mov     [esi+ecx*4],eax                 ; ...record
;PrintHex ecx,'cycle'
;PrintHex eax,'data'
                sub     ecx,1
                jnz @_21
                pop     ecx
                cmp     ecx,[edx].MAXPROBE              ; current number of probes greater?
                jc  @_3
                mov     [edx].MAXPROBE,ecx              ; store new number of probes
@_3:
                sub     [edx].FREE,1
                ASSUME edx:NOTHING
@_4:
                ret

HASH1_store     ENDP

; 

; 
;
; Searches for data in HASH
;
; IN:       hd - pointer to a HASH structure
;          eax - key of record
;
; OUT:  esi - PTR to found slot
;           - NULL when not found
;
; -------------------------------------------------------------------------
HASH1_search    PROC USES ecx edi hd:DWORD
align 4
                mov     edx,hd
                ASSUME edx:PTR HASH1_STRUCT
                mov     ecx,[edx].MAXPROBE
                call    DWORD PTR [edx].GETKEY          ; hash function: ebx - key/index
align 4
@_2:
                mov     esi,[edx].START
                push    eax
                push    edx
                mov     eax,[edx].rSIZE
                add     eax,1
                shl     eax,2
                mul     ebx
                pop     edx
                add     esi,eax
                pop     eax
;PrintHex esi
                mov     edi,[esi]                       ; read key
                or      edi,edi                         ; is empty?
                jz @_3
                cmp     edi,eax                         ; key found?
                jz @_4
                call    DWORD PTR [edx].PROBE           ; hash function: ebx - key/index
;PrintHex ecx,'probe'
                sub     ecx,1                           ; number of probes
                jnc  @_2
                ASSUME edx:NOTHING
; ----------------
; Empty...
@_3:
                xor     esi,esi
                ret
; ----------------
; Found...
@_4:
                add     esi,4
                ret

HASH1_search    ENDP

; 


end